proc MGCD_LD1(list F, int k, ideal Equ, ideal DisEqu, RingAll, RingVar)
{
  int s = k;
  export(RingVar, RingAll, s);
  setring(RingAll);

  list G = MGCD_LD1_main(F, Equ, DisEqu);
  keepring(RingAll);
  return(G);

}



static proc MGCD_LD1_main(list F, ideal Equ, ideal DisEqu)
{
  list Gcd = pcgs_main(F[1], Equ, DisEqu);
  if(size(F)>1)
  {
    int i,j;
    list G,G1;
    module M;
    for(i=2;i<=size(F);i++)
    {
      G = list();
      for(j=1;j<=size(Gcd);j++)
      { 
        if(Gcd[j][3][1] == 0)
        {
          G = G + pcgs_main(F[i], Gcd[j][1], Gcd[j][2]);
        }
        else
        {
          M = gen(1)+ gen(2) +gen(3), Gcd[j][3][1]*gen(1), F[i]*gen(2);
          G1 = mcgs_main(M, Gcd[j][1], Gcd[j][2]);
          G = G + sub_gcd1(Gcd[j][3][1]*F[i],G1);
        }      
      }
      Gcd = G;
    }
  }
  return(Gcd);
}          

 


static proc sub_gcd1(poly f, list G)
{
  int i,j;
  poly q;
  vector V = gen(1)+ gen(2) +gen(3);
  for(i=1;i<=size(G);i++)
  {
    if(size(G[i][3])==2)
    {
      if(G[i][3][1] != V)
      {
        G[i][3] = list(G[i][3][1][2]);
      }
      else
      {
        G[i][3] = list(G[i][3][2][2]);
      }
    }
    else
    {
      for(j=1;j<=size(G[i][3]);j++)
      {
        if(G[i][3][j][1]==0 && G[i][3][j][2]==0)
        {
          q = Parametric_Division(f, G[i][3][j][3], G[i][1], G[i][2]);
          G[i][3] = list(q);
          break;
        }
      }
    }
  }
  return(G);
}



static proc pcgs_main(ideal I, ideal Equ, ideal DisEqu)
{
  //check consistence of (Equ, DisEqu)
  if (!Check_Consistence(Equ, DisEqu))
  {
    return(list());
  }  

  ideal newEqu, newDisEqu;
  if(size(Equ) != 0)
  {
    newEqu = radical(Equ);
    option(redSB);
    newEqu = std(newEqu);
  }  
  if(DisEqu[1]==1)
  {
    newDisEqu = DisEqu;
  }
  else
  {
    newDisEqu = Simplify_DisEqu(newEqu,DisEqu);
  }

  //compute the full elinimating Groebner basis
  option(redSB);
  ideal GB = std(I + newEqu);
  if (GB[1] == 1)
  {
    return(list(list(newEqu, newDisEqu, list(1))));
  }

  //get minimal Dickson basis
  poly H;
  list MDBasis, CGSystems;
  ideal Gs, Gr, LeadCoeffFactors;
  MDBasis, Gs, Gr, H, LeadCoeffFactors = Separate_Polys(GB);

  //trivial branches
  if (size(Gr))
  {
    ideal Tempdisequ = newDisEqu * Gr;
    if (Check_Consistence(newEqu, Tempdisequ))
    {
      CGSystems = list(list(newEqu, Tempdisequ, list(1)));
    }
    if (!Check_Consistence(Gr, newDisEqu))
    {
      return(CGSystems);
    }
    if (size(MDBasis) == 0)
    {
      CGSystems = CGSystems + list(list(Gr, newDisEqu, list(0)));
      return(CGSystems);
    }
  }

  //major branch
  ideal tempp = newDisEqu * H;
  if (Check_Consistence(Gr, tempp))
  {
    CGSystems = CGSystems + list(list(Gr, tempp, MDBasis));
  }
  
  //other branches
  ideal tempDisEqu = newDisEqu;
  for (int i = 1; i <= size(LeadCoeffFactors); i++)
  {
    CGSystems = CGSystems + pcgs_main(Gs,Gr + LeadCoeffFactors[i], tempDisEqu);
    tempDisEqu = tempDisEqu * LeadCoeffFactors[i];
  }

  return(CGSystems);
}



static proc Separate_Polys(ideal GB)
{
  setring(RingVar);
  ideal newGB = imap(RingAll, GB);
  list MDBasis, GrList, MList;
  ideal Gs;
  poly P, mon;
  poly LCoefs = 1;
  int length = size(newGB);
  int i, j;
  for (i = 1; i <= length; i++)
  {
    P = newGB[i];
    mon = leadmonom(P);
    if (mon == 1)
    {
      GrList = GrList + list(P);
    }
    else
    {
      Gs = Gs + newGB[i];
      for (j = 1; j <= size(MList); j++)
      {
        if (mon/MList[j] > 1|| mon/MList[j] ==1)
        {
	  mon = 0;
	  break;
        }
      }
      if (mon == 0)
      {
        i++;
        continue;
      }
      for (j = i + 1; j <= length; j++)
      {
	if (leadmonom(newGB[j]) != 1 && mon/leadmonom(newGB[j]) > 1)
	{
	  mon = 0;
	  break;
	}
      }
      if (mon == 0)
      {
        i ++;
        continue;
      }
      MDBasis = MDBasis + list(P);
      MList = MList + list(mon);
      LCoefs = LCoefs * leadcoef(P);
    }
  }

  int MDBasisLength = size(MDBasis);
  int GrListLength = size(GrList);
  int GsListLength = size(Gs);

  setring(RingAll);

  list MDBasis;
  ideal Gs, Gr, LeadCoeffFactors;
  poly g;

  if (GsListLength)
  {
    Gs = imap(RingVar, Gs);
  }
  if (GrListLength)
  {
    list GrList = imap(RingVar, GrList);
    for (i = 1; i <= GrListLength; i++)
    {
      Gr = Gr + GrList[i];
    }
  }

  if (MDBasisLength)
  {
    MDBasis  =  imap(RingVar, MDBasis);
    g  =  imap(RingVar, LCoefs);
    g = sqrfree(g,3);
    LeadCoeffFactors = factorize(g,1);
  }

  return(MDBasis, Gs, Gr, g, LeadCoeffFactors);
}



static proc Check_Consistence(ideal Equ, ideal DisEqu)
{
  //trivial checks
  if (size(DisEqu) == 0 || Equ[1] == 1)
  {
    return(0);
  }
  if (size(Equ) == 0 || DisEqu[1] == 1)
  {
    return(1);
  }

  //normal check
  int i, flag;
  for (i=1;i<=size(DisEqu);i++)
  {
    flag = rad_con(DisEqu[i], Equ);
    if ( flag == 0 )
    {
      return(1);
    }
  }

  return(0);
}



static proc Simplify_DisEqu(ideal Equ, ideal DisEqu)
{
  ideal newDisEqu;
  if (DisEqu[1] == 1)
  {
    return(DisEqu);
  }
  if (size(Equ) == 0)
  {
    newDisEqu = radical(DisEqu);
    option(redSB);
    newDisEqu = std(newDisEqu);
    return(newDisEqu);
  }
  poly tempP;
  for (int i = 1; i <= size(DisEqu); i++)
  {
    tempP = reduce(DisEqu[i],Equ);
    if (leadmonom(tempP) == 1)
    {
      return(ideal(1));
    }
    else
    {
      if (tempP != 0)
      {
        newDisEqu = newDisEqu + tempP;
      }
    }
  }
  newDisEqu = radical(newDisEqu);
  option(redSB);
  newDisEqu = std(newDisEqu);
  return(newDisEqu);
}



static proc mcgs_main(module I, ideal Equ, ideal DisEqu)
{
  //check consistence of (Equ, DisEqu)
  if (!Check_Consistence(Equ, DisEqu))
  {
    return(list());
  }

  ideal newEqu, newDisEqu;
  if(size(Equ) != 0)
  {
    newEqu = radical(Equ);
    option(redSB);
    newEqu = std(newEqu);
  }  
  if(DisEqu[1]==1)
  {
    newDisEqu = DisEqu;
  }
  else
  {
    newDisEqu = Simplify_DisEqu(newEqu,DisEqu);
  }

  //generate the standard basis of k[U][X]^s and transform newEqu into k[U]^s
  list T;
  module NewEqu;
  for(int i=1;i<=s;i++)
  {
    T = T + list(gen(i));
    NewEqu = NewEqu + newEqu * gen(i);
  }

  //compute the full elinimating Grobner basis
  option(redSB);
  module GB = std(I + NewEqu);
  int k = UnitInModule(GB);
  if (k)
  {
    return(list(list(newEqu, newDisEqu, T)));
  }

  //get minimal Dickson basis
  poly H;
  module Gs;
  list MDBasis, CGSystems;
  ideal Gr, LeadCoeffFactors;
  MDBasis, Gs, Gr, H, LeadCoeffFactors = Separate_Modules(GB);

  //trivial branches
  if (size(Gr))
  {
    ideal Tempdisequ = newDisEqu * Gr;
    if (Check_Consistence(newEqu, Tempdisequ))
    {
      CGSystems = list(list(newEqu, Tempdisequ, T));
    }
    if (!Check_Consistence(Gr, newDisEqu))
    {
      return(CGSystems);
    }
    if (size(MDBasis) == 0)
    {
      vector v;
      CGSystems = CGSystems + list(list(Gr, newDisEqu, list(v)));
      return(CGSystems);
    }
  }
  
  //major branch
  ideal tempp = newDisEqu * H;
  if (Check_Consistence(Gr, tempp))
  {
    CGSystems = CGSystems + list(list(Gr, tempp, MDBasis));
  }
  
  //other branches
  ideal tempDisEqu = newDisEqu;
  for (i = 1; i <= size(LeadCoeffFactors); i++)
  {
    CGSystems = CGSystems + mcgs_main(Gs,Gr + LeadCoeffFactors[i], tempDisEqu);
    tempDisEqu = tempDisEqu * LeadCoeffFactors[i];
  }

  return(CGSystems);
}


static proc Separate_Modules(module GB)
{
  setring(RingVar);
  module newGB = imap(RingAll, GB);
  list MDBasis, GrList, MList;
  module Gs;
  poly mon,Mon;
  poly LCoefs = 1;
  int length = size(newGB);
  int i, j, r1, k, r2, flag;
  for (i = 1; i <= length; i++)
  {
    r1, k, mon = LeadInVector(newGB[i]);
    if (r1 == 1 && mon == 1)
    {
      GrList = GrList + list(newGB[i][k]);
    }
    else
    {
      Gs = Gs + newGB[i];
      for (j = 1; j <= size(MList); j++)
      {
        if (k == MList[j][1])
        {
          if (mon/MList[j][2] > 1 || mon/MList[j][2] ==1)
          {
	    mon = 0;
	    break;
          }
        }
      }
      if (mon == 0)
      {
        i++;
        continue;
      }
      for (j = i + 1; j <= length; j++)
      {
        r2, flag, Mon = LeadInVector(newGB[j]);
        if (k == flag)
        {
	  if (r2 == 0 && mon/Mon > 1)
	  {
	    mon = 0;
	    break;
	  }
        }
      }
      if (mon == 0)
      {
        i ++;
        continue;
      }
      MDBasis = MDBasis + list(newGB[i]);
      MList = MList + list(list(k,mon));
      LCoefs = LCoefs * leadcoef(newGB[i][k]);
    }
  }

  int MDBasisLength = size(MDBasis);
  int GrListLength = size(GrList);
  int GsListLength = size(Gs);

  setring(RingAll);

  list MDBasis;
  module Gs;
  ideal Gr, LeadCoeffFactors;
  poly g;

  if (GsListLength)
  {
    Gs = imap(RingVar, Gs);
  }
  if (GrListLength)
  {
    list GrList = imap(RingVar, GrList);
    for (i = 1; i <= GrListLength; i++)
    {
      Gr = Gr + GrList[i];
    }
  }

  if (MDBasisLength)
  {
    MDBasis  =  imap(RingVar, MDBasis);
    g  =  imap(RingVar, LCoefs);
    g = sqrfree(g,3);
    LeadCoeffFactors = factorize(g,1);
  }

  return(MDBasis, Gs, Gr, g, LeadCoeffFactors);
}



static proc UnitInModule(module M)
{
  int i;
  int flag = 1;
  vector v;
  for(i=1;i<=s;i++)
  {
    v = reduce(gen(i),M);
    if (v != 0)
    {
      flag = 0;
      break;
    }
  }
  return(flag);
}



static proc LeadInVector(vector V)
{
  vector Monom = leadmonom(V);
  int i,flag;
  vector W;
  for(i=1;i<=s;i++)
  {
    if(V == gen(i))
    {
      return(0,i,1);
    }
    else
    {
      if(Monom[i] != 0)
      {
        W = V - leadcoef(V)*gen(i);
        if(W == 0)
        {
          flag = 1;
        }
        return(flag,i,Monom[i]);
      }
    }
  }
}



static proc Poly_Simplify(poly h, ideal Equ, ideal DisEqu)
{
  if(DisEqu[1]==1)
  {
    ideal G = radical(Equ);
    option(redSB);
    G = std(G);
  }
  else
  {
    ideal G = quotient(Equ,DisEqu);
    G = radical(G);
    option(redSB);
    G = std(G);
  }

  setring(RingVar);
  poly h1 = imap(RingAll,h);
  list T1;
  while (size(h1) > 0)
  {  
    T1 = T1 + list(list(leadmonom(h1),leadcoef(h1)));
    h1 = h1 - lead(h1);
  } 
  setring(RingAll);
  list T2 = imap(RingVar,T1);
  poly r,h2,h3;
  for (int i=1;i<=size(T2);i++)
  {
    r = reduce(T2[i][2],G);
    h2 = h2 + T2[i][1]*r;
    h3 = gcd(h3,r);
  }
  h2 = h2/h3;
  h2 = h2/content(h2);
  return(h2);
}



static proc Parametric_Division(poly f, poly g, ideal Equ, ideal DisEqu)
{
  setring(RingVar);
  poly f1 = imap(RingAll,f);
  ideal I = imap(RingAll,g);
  list S = pdivi(f1,I);
  poly q = S[2][1];
  poly mon = leadmonom(q);
  int flag;
  if(mon ==1)
  {
    flag = 1;
  }
  setring(RingAll);
  if(flag)
  {
    poly q = 1;
    return(q);
  }
  poly q = imap(RingVar,q);
  if(size(Equ) != 0)
  {
    q = Poly_Simplify(q, Equ, DisEqu);
  }
  return(q);
}



proc StringGCD(list cgs)
{
  int i, j, k;
  string Str = newline + "The gcd system of parametric polynomials on V(E)-V(N) are:" + newline + newline;
  for (i = 1; i <= size(cgs); i++)
  {
    Str = Str + "Branch " + string(i) + ":" + newline + "parametric constraint V(E): " + newline;
    k = size(cgs[i][1]);
    if ( k == 0 )
    { 
      Str = Str + "V(0)";
    }
    else
    {
      Str = Str + "V(";
      if(k==1)
      {
        Str = Str + string(cgs[i][1][1]) + ")";
      }
      else
      {
        for (j = 1; j <= k-1; j++)
        {
          Str = Str + string(cgs[i][1][j]) + ", ";
        }
        Str = Str + string(cgs[i][1][k]) + ")";
      }
    }
    Str = Str + newline + "parametric constraint V(N): " + newline;
    k = size(cgs[i][2]);
    if ( k == 0 )
    { 
      Str = Str + "V(1)" ;
    }
    else
    {
      Str = Str + "V(";
      if(k==1)
      {
        Str = Str + string(cgs[i][2][1]) + ")";
      }
      else
      {
        for (j = 1; j <= k-1; j++)
        {
          Str = Str + string(cgs[i][2][j]) + ", ";
        }
        Str = Str + string(cgs[i][2][k]) + ")";
      }
    }
    Str = Str + newline + "The gcd of parametric polynomials on V(E)-V(N) is: " + newline;
    Str = Str + string(cgs[i][3][1]) + "." + newline + newline;
  }
  return(Str);
}