proc MGCD_LD3(list F, ideal Equ, ideal DisEqu, RingAll, RingVar)
{
  export(RingVar, RingAll);
  setring(RingAll);
  list G = MGCD_LD3_main(F, Equ, DisEqu);
  keepring(RingAll);
  return(G);
}



static proc MGCD_LD3_main(list F, ideal Equ, ideal DisEqu)
{
  list Gcd = pcgs_main(F[1], Equ, DisEqu);
  if(size(F)>1)
  {
    list G,G1;
    module M;
    list T = F[2..size(F)];
    for(int j=1;j<=size(Gcd);j++)
    { 
      if(Gcd[j][3][1] == 0)
      {
        G = G + MGCD_LD3(T, Gcd[j][1], Gcd[j][2], RingAll, RingVar);
      }
      else
      {
        int s = 2 * size(F) - 1;
        M = Syzmodule(Gcd[j][3][1], T);
        G1 = mcgs(M, s, Gcd[j][1], Gcd[j][2]);
        G = G + sub_gcd3(Gcd[j][3][1],size(F),G1);
      }
    }
    Gcd = G;
  }
  return(Gcd);
}          



static proc Syzmodule(poly f, list F)
{
  int k = size(F);
  module M;
  int i;
  for(i=1;i<=k;i++)
  {
    M[1] = M[1] + F[i]*gen(i);
    M[i+1] = f * gen(i) + gen(k+1+i);
  }
  M[1] = M[1] + gen(k+1);
  return(M);
}



static proc sub_gcd3(poly f, int n, list G)
{
  int i, j, k, flag;
  poly q;
  for(i=1;i<=size(G);i++)
  {
    for(j=1;j<=size(G[i][3]);j++)
    {
      flag = 1;
      for(k=1;k<=n-1;k++)
      {
        if(G[i][3][j][k] != 0)
        {
          flag = 0;
          break;
        }
      }
      if(flag)
      {
        q = Parametric_Division(f, G[i][3][j][n], G[i][1], G[i][2]);
        G[i][3] = list(q);
        break;
      }
    }
  }
  return(G);
}



static proc Poly_Simplify(poly h, ideal Equ, ideal DisEqu)
{
  if(DisEqu[1]==1)
  {
    ideal G = radical(Equ);
    option(redSB);
    G = std(G);
  }
  else
  {
    ideal G = quotient(Equ,DisEqu);
    G = radical(G);
    option(redSB);
    G = std(G);
  }
  setring(RingVar);
  poly h1 = imap(RingAll,h);
  list T1;
  while (size(h1) > 0)
  {  
    T1 = T1 + list(list(leadmonom(h1),leadcoef(h1)));
    h1 = h1 - lead(h1);
  } 
  setring(RingAll);
  list T2 = imap(RingVar,T1);
  poly r,h2,h3;
  for (int i=1;i<=size(T2);i++)
  {
    r = reduce(T2[i][2],G);
    h2 = h2 + T2[i][1]*r;
    h3 = gcd(h3,r);
  }
  h2 = h2/h3;
  h2 = h2/content(h2);
  return(h2);
}



static proc Parametric_Division(poly f, poly g, ideal Equ, ideal DisEqu)
{
  setring(RingVar);
  poly f1 = imap(RingAll,f);
  ideal I = imap(RingAll,g);
  list S = pdivi(f1,I);
  poly q = S[2][1];
  poly mon = leadmonom(q);
  int flag;
  if(mon ==1)
  {
    flag = 1;
  }
  setring(RingAll);
  if(flag)
  {
    poly q = 1;
    return(q);
  }
  poly q = imap(RingVar,q);
  if(size(Equ) != 0)
  {
    q = Poly_Simplify(q, Equ, DisEqu);
  }
  return(q);
}



proc StringGCD(list cgs)
{
  int i, j, k;
  string Str = newline + "The gcd system of parametric polynomials on V(E)-V(N) are:" + newline + newline;
  for (i = 1; i <= size(cgs); i++)
  {
    Str = Str + "Branch " + string(i) + ":" + newline + "parametric constraint V(E): " + newline;
    k = size(cgs[i][1]);
    if ( k == 0 )
    { 
      Str = Str + "V(0)";
    }
    else
    {
      Str = Str + "V(";
      if(k==1)
      {
        Str = Str + string(cgs[i][1][1]) + ")";
      }
      else
      {
        for (j = 1; j <= k-1; j++)
        {
          Str = Str + string(cgs[i][1][j]) + ", ";
        }
        Str = Str + string(cgs[i][1][k]) + ")";
      }
    }
    Str = Str + newline + "parametric constraint V(N): " + newline;
    k = size(cgs[i][2]);
    if ( k == 0 )
    { 
      Str = Str + "V(1)" ;
    }
    else
    {
      Str = Str + "V(";
      if(k==1)
      {
        Str = Str + string(cgs[i][2][1]) + ")";
      }
      else
      {
        for (j = 1; j <= k-1; j++)
        {
          Str = Str + string(cgs[i][2][j]) + ", ";
        }
        Str = Str + string(cgs[i][2][k]) + ")";
      }
    }
    Str = Str + newline + "The gcd of parametric polynomials on V(E)-V(N) is: " + newline;
    Str = Str + string(cgs[i][3][1]) + "." + newline + newline;
  }
  return(Str);
}
