//**************************************

// CGS for parametric polynomials

proc pcgs_main(ideal I, ideal Equ, ideal DisEqu)
{
  //check consistence of (Equ, DisEqu)
  if (!Check_Consistence(Equ, DisEqu))
  {
    return(list());
  }
  ideal newEqu, newDisEqu;
  if(size(Equ) != 0)
  {
    newEqu = radical(Equ);
    option(redSB);
    newEqu = std(newEqu);
  }  
  if(DisEqu[1]==1)
  {
    newDisEqu = DisEqu;
  }
  else
  {
    newDisEqu = Simplify_DisEqu(newEqu,DisEqu);
  }

  //compute the full elinimating Groebner basis
  option(redSB);
  ideal GB = std(I + newEqu);
  if (GB[1] == 1)
  {
    return(list(list(newEqu, newDisEqu, list(1))));
  }

  //get minimal Dickson basis
  poly H;
  list MDBasis, CGSystems;
  ideal Gs, Gr, LeadCoeffFactors;
  MDBasis, Gs, Gr, H, LeadCoeffFactors = Separate_Polys(GB);

  //trivial branches
  if (size(Gr))
  {
    ideal Tempdisequ = newDisEqu * Gr;
    if (Check_Consistence(newEqu, Tempdisequ))
    {
      CGSystems = list(list(newEqu, Tempdisequ, list(1)));
    }
    if (!Check_Consistence(Gr, newDisEqu))
    {
      return(CGSystems);
    }
    if (size(MDBasis) == 0)
    {
      CGSystems = CGSystems + list(list(Gr, newDisEqu, list(0)));
      return(CGSystems);
    }
  }

  //major branch
  ideal tempp = newDisEqu * H;
  if (Check_Consistence(Gr, tempp))
  {
    CGSystems = CGSystems + list(list(Gr, tempp, MDBasis));
  }
  
  //other branches
  ideal tempDisEqu = newDisEqu;
  for (int i = 1; i <= size(LeadCoeffFactors); i++)
  {
    CGSystems = CGSystems + pcgs_main(Gs,Gr + LeadCoeffFactors[i], tempDisEqu);
    tempDisEqu = tempDisEqu * LeadCoeffFactors[i];
  }

  return(CGSystems);
}



// compute the minimal Dickson basis of polynomials

static proc Separate_Polys(ideal GB)
{
  setring(RingVar);
  ideal newGB = imap(RingAll, GB);
  list MDBasis, GrList, MList;
  ideal Gs;
  poly P, mon;
  poly LCoefs = 1;
  int length = size(newGB);
  int i, j;
  for (i = 1; i <= length; i++)
  {
    P = newGB[i];
    mon = leadmonom(P);
    if (mon == 1)
    {
      GrList = GrList + list(P);
    }
    else
    {
      Gs = Gs + newGB[i];
      for (j = 1; j <= size(MList); j++)
      {
        if (mon/MList[j] > 1|| mon/MList[j] ==1)
        {
	  mon = 0;
	  break;
        }
      }
      if (mon == 0)
      {
        i++;
        continue;
      }
      for (j = i + 1; j <= length; j++)
      {
	if (leadmonom(newGB[j]) != 1 && mon/leadmonom(newGB[j]) > 1)
	{
	  mon = 0;
	  break;
	}
      }
      if (mon == 0)
      {
        i ++;
        continue;
      }
      MDBasis = MDBasis + list(P);
      MList = MList + list(mon);
      LCoefs = LCoefs * leadcoef(P);
    }
  }
  int MDBasisLength = size(MDBasis);
  int GrListLength = size(GrList);
  int GsListLength = size(Gs);

  setring(RingAll);
  list MDBasis;
  ideal Gs, Gr, LeadCoeffFactors;
  poly g;

  if (GsListLength)
  {
    Gs = imap(RingVar, Gs);
  }
  if (GrListLength)
  {
    list GrList = imap(RingVar, GrList);
    for (i = 1; i <= GrListLength; i++)
    {
      Gr = Gr + GrList[i];
    }
  }
  if (MDBasisLength)
  {
    MDBasis  =  imap(RingVar, MDBasis);
    g  =  imap(RingVar, LCoefs);
    g = sqrfree(g,3);
    LeadCoeffFactors = factorize(g,1);
  }

  return(MDBasis, Gs, Gr, g, LeadCoeffFactors);
}



//**************************************

// CGS for parametric modules

proc mcgs(module I, int k, ideal Equ, ideal DisEqu)
{
  int s = k;
  export(s);
  setring(RingAll);
  list G = mcgs_main(I, Equ, DisEqu);
  keepring(RingAll);
  return(G);
}



static proc mcgs_main(module I, ideal Equ, ideal DisEqu)
{
  //check consistence of (Equ, DisEqu)
  if (!Check_Consistence(Equ, DisEqu))
  {
    return(list());
  }

  ideal newEqu, newDisEqu;
  if(size(Equ) != 0)
  {
    newEqu = radical(Equ);
    option(redSB);
    newEqu = std(newEqu);
  }  
  if(DisEqu[1]==1)
  {
    newDisEqu = DisEqu;
  }
  else
  {
    newDisEqu = Simplify_DisEqu(newEqu,DisEqu);
  }

  //generate the standard basis of k[U][X]^s and transform newEqu into k[U]^s
  list T;
  module NewEqu;
  for(int i=1;i<=s;i++)
  {
    T = T + list(gen(i));
    NewEqu = NewEqu + newEqu * gen(i);
  }

  //compute the full elinimating Grobner basis
  option(redSB);
  module GB = std(I + NewEqu);
  int k = UnitInModule(GB);
  if (k)
  {
    return(list(list(newEqu, newDisEqu, T)));
  }

  //get minimal Dickson basis
  poly H;
  module Gs;
  list MDBasis, CGSystems;
  ideal Gr, LeadCoeffFactors;
  MDBasis, Gs, Gr, H, LeadCoeffFactors = Separate_Modules(GB);

  //trivial branches
  if (size(Gr))
  {
    ideal Tempdisequ = newDisEqu * Gr;
    if (Check_Consistence(newEqu, Tempdisequ))
    {
      CGSystems = list(list(newEqu, Tempdisequ, T));
    }
    if (!Check_Consistence(Gr, newDisEqu))
    {
      return(CGSystems);
    }
    if (size(MDBasis) == 0)
    {
      vector v;
      CGSystems = CGSystems + list(list(Gr, newDisEqu, list(v)));
      return(CGSystems);
    }
  }
  
  //major branch
  ideal tempp = newDisEqu * H;
  if (Check_Consistence(Gr, tempp))
  {
    CGSystems = CGSystems + list(list(Gr, tempp, MDBasis));
  }
  
  //other branches
  ideal tempDisEqu = newDisEqu;
  for (i = 1; i <= size(LeadCoeffFactors); i++)
  {
    CGSystems = CGSystems + mcgs_main(Gs,Gr + LeadCoeffFactors[i], tempDisEqu);
    tempDisEqu = tempDisEqu * LeadCoeffFactors[i];
  }

  return(CGSystems);
}



// compute the minimal Dickson basis for modules

static proc Separate_Modules(module GB)
{
  setring(RingVar);
  module newGB = imap(RingAll, GB);
  list MDBasis, GrList, MList;
  module Gs;
  poly mon,Mon;
  poly LCoefs = 1;
  int length = size(newGB);
  int i, j, r1, k, r2, flag;
  for (i = 1; i <= length; i++)
  {
    r1, k, mon = LeadInVector(newGB[i]);
    if (r1 == 1 && mon == 1)
    {
      GrList = GrList + list(newGB[i][k]);
    }
    else
    {
      Gs = Gs + newGB[i];
      for (j = 1; j <= size(MList); j++)
      {
        if (k == MList[j][1])
        {
          if (mon/MList[j][2] > 1 || mon/MList[j][2] ==1)
          {
	    mon = 0;
	    break;
          }
        }
      }
      if (mon == 0)
      {
        i++;
        continue;
      }
      for (j = i + 1; j <= length; j++)
      {
        r2, flag, Mon = LeadInVector(newGB[j]);
        if (k == flag)
        {
	  if (r2 == 0 && mon/Mon > 1)
	  {
	    mon = 0;
	    break;
	  }
        }
      }
      if (mon == 0)
      {
        i ++;
        continue;
      }
      MDBasis = MDBasis + list(newGB[i]);
      MList = MList + list(list(k,mon));
      LCoefs = LCoefs * leadcoef(newGB[i][k]);
    }
  }

  int MDBasisLength = size(MDBasis);
  int GrListLength = size(GrList);
  int GsListLength = size(Gs);

  setring(RingAll);

  list MDBasis;
  module Gs;
  ideal Gr, LeadCoeffFactors;
  poly g;

  if (GsListLength)
  {
    Gs = imap(RingVar, Gs);
  }
  if (GrListLength)
  {
    list GrList = imap(RingVar, GrList);
    for (i = 1; i <= GrListLength; i++)
    {
      Gr = Gr + GrList[i];
    }
  }

  if (MDBasisLength)
  {
    MDBasis  =  imap(RingVar, MDBasis);
    g  =  imap(RingVar, LCoefs);
    g = sqrfree(g,3);
    LeadCoeffFactors = factorize(g,1);
  }

  return(MDBasis, Gs, Gr, g, LeadCoeffFactors);
}



static proc UnitInModule(module M)
{
  int i;
  int flag = 1;
  vector v;
  for(i=1;i<=s;i++)
  {
    v = reduce(gen(i),M);
    if (v != 0)
    {
      flag = 0;
      break;
    }
  }
  return(flag);
}


static proc LeadInVector(vector V)
{
  vector Monom = leadmonom(V);
  int i,flag;
  vector W;
  for(i=1;i<=s;i++)
  {
    if(V == gen(i))
    {
      return(0,i,1);
    }
    else
    {
      if(Monom[i] != 0)
      {
        W = V - leadcoef(V)*gen(i);
        if(W == 0)
        {
          flag = 1;
        }
        return(flag,i,Monom[i]);
      }
    }
  }
}



//**************************************

static proc Check_Consistence(ideal Equ, ideal DisEqu)
{
  //trivial checks
  if (size(DisEqu) == 0 || Equ[1] == 1)
  {
    return(0);
  }
  if (size(Equ) == 0 || DisEqu[1] == 1)
  {
    return(1);
  }

  //normal check
  int i, flag;
  for (i=1;i<=size(DisEqu);i++)
  {
    flag = rad_con(DisEqu[i], Equ);
    if ( flag == 0 )
    {
      return(1);
    }
  }

  return(0);
}



static proc Simplify_DisEqu(ideal Equ, ideal DisEqu)
{
  ideal newDisEqu;
  if (DisEqu[1] == 1)
  {
    return(DisEqu);
  }
  if (size(Equ) == 0)
  {
    newDisEqu = radical(DisEqu);
    option(redSB);
    newDisEqu = std(newDisEqu);
    return(newDisEqu);
  }
  poly tempP;
  for (int i = 1; i <= size(DisEqu); i++)
  {
    tempP = reduce(DisEqu[i],Equ);
    if (leadmonom(tempP) == 1)
    {
      return(ideal(1));
    }
    else
    {
      if (tempP != 0)
      {
        newDisEqu = newDisEqu + tempP;
      }
    }
  }
  newDisEqu = radical(newDisEqu);
  option(redSB);
  newDisEqu = std(newDisEqu);
  return(newDisEqu);
}