proc pcgs(ideal I, ideal Equ, ideal DisEqu, RingAll, RingVar)
{
  export(RingVar, RingAll);
  setring(RingAll);
  list G = pcgs_main(I, Equ, DisEqu);
  keepring(RingAll);
  return(G);
}


static proc pcgs_main(ideal I, ideal Equ, ideal DisEqu)
{
  //check consistence of (Equ, DisEqu)
  if (!Check_Consistence(Equ, DisEqu))
  {
    return(list());
  }  

  ideal newEqu, newDisEqu;
  if(size(Equ) != 0)
  {
    newEqu = radical(Equ);
    option(redSB);
    newEqu = std(newEqu);
  }  
  if(DisEqu[1]==1)
  {
    newDisEqu = DisEqu;
  }
  else
  {
    newDisEqu = Simplify_DisEqu(newEqu,DisEqu);
  }

  //compute the full elinimating Groebner basis
  option(redSB);
  ideal GB = std(I + newEqu);
  if (GB[1] == 1)
  {
    return(list(list(newEqu, newDisEqu, list(1))));
  }

  //get minimal Dickson basis
  poly H;
  list MDBasis, CGSystems;
  ideal Gs, Gr, LeadCoeffFactors;
  MDBasis, Gs, Gr, H, LeadCoeffFactors = Separate_Polys(GB);

  //trivial branches
  if (size(Gr))
  {
    ideal Tempdisequ = newDisEqu * Gr;
    if (Check_Consistence(newEqu, Tempdisequ))
    {
      CGSystems = list(list(newEqu, Tempdisequ, list(1)));
    }
    if (!Check_Consistence(Gr, newDisEqu))
    {
      return(CGSystems);
    }
    if (size(MDBasis) == 0)
    {
      CGSystems = CGSystems + list(list(Gr, newDisEqu, list(0)));
      return(CGSystems);
    }
  }

  //major branch
  ideal tempp = newDisEqu * H;
  if (Check_Consistence(Gr, tempp))
  {
    CGSystems = CGSystems + list(list(Gr, tempp, MDBasis));
  }
  
  //other branches
  ideal tempDisEqu = newDisEqu;
  for (int i = 1; i <= size(LeadCoeffFactors); i++)
  {
    CGSystems = CGSystems + pcgs_main(Gs,Gr + LeadCoeffFactors[i], tempDisEqu);
    tempDisEqu = tempDisEqu * LeadCoeffFactors[i];
  }

  return(CGSystems);
}


static proc Separate_Polys(ideal GB)
{
  setring(RingVar);
  ideal newGB = imap(RingAll, GB);
  list MDBasis, GrList, MList;
  ideal Gs;
  poly P, mon;
  poly LCoefs = 1;
  int length = size(newGB);
  int i, j;
  for (i = 1; i <= length; i++)
  {
    P = newGB[i];
    mon = leadmonom(P);
    if (mon == 1)
    {
      GrList = GrList + list(P);
    }
    else
    {
      Gs = Gs + newGB[i];
      for (j = 1; j <= size(MList); j++)
      {
        if (mon/MList[j] > 1|| mon/MList[j] ==1)
        {
	  mon = 0;
	  break;
        }
      }
      if (mon == 0)
      {
        i++;
        continue;
      }
      for (j = i + 1; j <= length; j++)
      {
	if (leadmonom(newGB[j]) != 1 && mon/leadmonom(newGB[j]) > 1)
	{
	  mon = 0;
	  break;
	}
      }
      if (mon == 0)
      {
        i ++;
        continue;
      }
      MDBasis = MDBasis + list(P);
      MList = MList + list(mon);
      LCoefs = LCoefs * leadcoef(P);
    }
  }

  int MDBasisLength = size(MDBasis);
  int GrListLength = size(GrList);
  int GsListLength = size(Gs);

  setring(RingAll);

  list MDBasis;
  ideal Gs, Gr, LeadCoeffFactors;
  poly g;

  if (GsListLength)
  {
    Gs = imap(RingVar, Gs);
  }
  if (GrListLength)
  {
    list GrList = imap(RingVar, GrList);
    for (i = 1; i <= GrListLength; i++)
    {
      Gr = Gr + GrList[i];
    }
  }

  if (MDBasisLength)
  {
    MDBasis  =  imap(RingVar, MDBasis);
    g  =  imap(RingVar, LCoefs);
    g = sqrfree(g,3);
    LeadCoeffFactors = factorize(g,1);
  }

  return(MDBasis, Gs, Gr, g, LeadCoeffFactors);
}



static proc Check_Consistence(ideal Equ, ideal DisEqu)
{
  //trivial checks
  if (size(DisEqu) == 0 || Equ[1] == 1)
  {
    return(0);
  }
  if (size(Equ) == 0 || DisEqu[1] == 1)
  {
    return(1);
  }

  //normal check
  int i, flag;
  for (i=1;i<=size(DisEqu);i++)
  {
    flag = rad_con(DisEqu[i], Equ);
    if ( flag == 0 )
    {
      return(1);
    }
  }

  return(0);
}



static proc Simplify_DisEqu(ideal Equ, ideal DisEqu)
{
  ideal newDisEqu;
  if (DisEqu[1] == 1)
  {
    return(DisEqu);
  }
  if (size(Equ) == 0)
  {
    newDisEqu = radical(DisEqu);
    option(redSB);
    newDisEqu = std(newDisEqu);
    return(newDisEqu);
  }
  poly tempP;
  for (int i = 1; i <= size(DisEqu); i++)
  {
    tempP = reduce(DisEqu[i],Equ);
    if (leadmonom(tempP) == 1)
    {
      return(ideal(1));
    }
    else
    {
      if (tempP != 0)
      {
        newDisEqu = newDisEqu + tempP;
      }
    }
  }
  newDisEqu = radical(newDisEqu);
  option(redSB);
  newDisEqu = std(newDisEqu);
  return(newDisEqu);
}



proc StringPCGS(list cgs)
{
  int i, j, k;
  
  string Str = newline + "The CGS for parametric polynomials on V(E)-V(N) is:" + newline + newline;
  for (i = 1; i <= size(cgs); i++)
  {
    Str = Str + "Branch " + string(i) + ":" + newline + "parametric constraint V(E): " + newline;
    k = size(cgs[i][1]);
    if ( k == 0 )
    { 
      Str = Str + "V(0)";
    }
    else
    {
      Str = Str + "V(";
      if(k==1)
      {
        Str = Str + string(cgs[i][1][1]) + ")";
      }
      else
      {
        for (j = 1; j <= k-1; j++)
        {
          Str = Str + string(cgs[i][1][j]) + ", ";
        }
        Str = Str + string(cgs[i][1][k]) + ")";
      }
    }

    Str = Str + newline + "parametric constraint V(N): " + newline;
    k = size(cgs[i][2]);
    if ( k == 0 )
    { 
      Str = Str + "V(1)" ;
    }
    else
    {
      Str = Str + "V(";
      if(k==1)
      {
        Str = Str + string(cgs[i][2][1]) + ")";
      }
      else
      {
        for (j = 1; j <= k-1; j++)
        {
          Str = Str + string(cgs[i][2][j]) + ", ";
        }
        Str = Str + string(cgs[i][2][k]) + ")";
      }
    }
    
    Str = Str + newline + "The Groebner basis for parametric polynomials on V(E)-V(N) is: " + newline;
    k = size(cgs[i][3]);
    if(k==1)
    {
      Str = Str + string(cgs[i][3][1]) + "." + newline + newline;
    }
    else
    {
      for (j = 1; j < k; j++)
      {
        Str = Str + string(cgs[i][3][j]) + ",";
      }
      Str = Str + string(cgs[i][3][k]) + "." + newline + newline;
    }
  }
  return(Str);
}